# test_methods.py
from pivpy import io, pivpy
import numpy as np
import pkg_resources as pkg
import pytest

import os

f1 = "Run000001.T000.D000.P000.H001.L.vec"
f2 = "Run000002.T000.D000.P000.H001.L.vec"
path = pkg.resource_filename("pivpy", "data/Insight")

_a = io.load_vec(os.path.join(path, f1))
_b = io.load_vec(os.path.join(path, f2))


def test_crop():
    """ tests crop """
    _c = _a.piv.crop([5, 15, -5, -15])
    assert _c.u.shape == (32, 32, 1)

    _c = io.create_sample_dataset()
    _c = _c.sel(x=slice(35, 70), y=slice(30, 90))
    assert _c.u.shape == (4, 1, 5)  # note the last dimension is preserved



def test_pan():
    """ test a shift by dx,dy using pan method """
    _a = io.load_vec(os.path.join(path, f1))
    _c = _a.piv.pan(1.0, -1.0)  # note the use of .piv.
    assert np.allclose(_c.coords["x"][0], 1.312480)
    assert np.allclose(_c.coords["y"][0], -1.31248)


def test_mean():
    data = io.create_sample_dataset(10)
    print(data.piv.average.u.median())
    assert np.allclose(data.piv.average.u.median(), 6.0)


def test_vec2scal():
    data = io.create_sample_dataset()
    data.piv.vec2scal()
    data.piv.vec2scal(property="curl")
    data.piv.vec2scal(property="ke")
    assert len(data.attrs["variables"]) == 5
    assert data.attrs["variables"][-1] == "ke"


def test_add():
    data = io.create_sample_dataset()
    tmp = data + data
    assert tmp["u"][0, 0, 0] == 2.0


def test_subtract():
    """ tests subtraction """
    data = io.create_sample_dataset()
    tmp = data - data
    assert tmp["u"][0, 0, 0] == 0.0


def test_multiply():
    """ tests subtraction """
    data = io.create_sample_dataset()
    tmp = data * 3.5
    assert tmp["u"][0, 0, 0] == 3.5


def test_set_get_dt():
    """ tests setting the new dt """
    data = io.create_sample_dataset()
    assert data.attrs["dt"] == 1.0
    assert data.piv.dt == 1.0
    data.piv.set_dt(2.0)
    assert data.attrs["dt"] == 2.0


# def test_rotate():
#     """ tests rotation """
#     data = io.create_sample_dataset()
#     data.piv.rotate(90) # rotate by 90 deg
#     assert data['u'][0,0,0] == 2.1 # shall fail

def test_fluctuations():
    data = io.create_sample_field()
    with pytest.raises(ValueError):
        data.piv.fluct()
    
    data = io.create_sample_dataset(100) # enough for random
    fluct = data.piv.fluct()
    assert np.allclose(fluct['u'].mean(dim='t'), 0.0)
    assert np.allclose(fluct['v'].mean(dim='t'), 0.0)


def test_reynolds_stress():
    data = io.create_sample_dataset(100)
    tmp = data.piv.reynolds_stress()
    assert np.allclose(tmp['w'].mean(dim='t'), 0.0)


def test_vorticity():
    """ tests vorticity estimate """
    data = io.create_sample_field() # we need another flow field
    data.piv.vorticity()
    assert np.allclose(data['w'], 0.0)


def test_strain():
    """ tests shear estimate """
    data = io.create_sample_field(noise_sigma=0.)
    data.piv.strain()
    assert np.allclose(data['w'], 0.00223713)


def test_tke():
    """ tests TKE """
    data = io.create_sample_dataset()
    data.piv.vec2scal(property="ke")
    data.piv.vec2scal(property="tke")  # now defined
    assert data.attrs["variables"][-1] == "tke"


def test_curl():
    """ tests curl that is also vorticity """
    _a = io.load_vec(os.path.join(path, f1))
    _a.piv.vec2scal(property="curl")
    assert _a.attrs["variables"][-1] == "vorticity"



import numpy as np
from pivpy import io, pivpy
import matplotlib.pyplot as plt

import os
import pkg_resources as pkg

path = pkg.resource_filename("pivpy", "data")
fname = os.path.join(path, "Insight", "Run000002.T000.D000.P000.H001.L.vec")

# def test_get_dt():
#     """ test if we get correct delta t """
#     _, _, _, _,dt,_ = io.parse_header(os.path.join(path,fname))
#     assert dt == 2000.


def test_get_frame():
    """ tests the correct frame number """
    _, _, _, _, _, frame = io.parse_header(
        os.path.join(path, "day2/day2a005003.T000.D000.P003.H001.L.vec")
    )
    assert frame == 5003
    _, _, _, _, _, frame = io.parse_header(
        os.path.join(path, "Insight/Run000002.T000.D000.P000.H001.L.vec")
    )
    assert frame == 2
    _, _, _, _, _, frame = io.parse_header(
        os.path.join(path, "openpiv/exp1_001_b.vec")
    )
    assert frame == 1
    _, _, _, _, _, frame = io.parse_header(
        os.path.join(path, "openpiv/exp1_001_b.txt")
    )
    assert frame == 1


def test_get_units():
    # test vec file with m/s
    lUnits, vUnits, tUnits = io.get_units(fname)
    assert lUnits == "mm"
    assert vUnits == "m/s"
    assert tUnits == "s"

    # test vec file with pixels/dt
    lUnits, vUnits, tUnits = io.get_units(
        os.path.join(path, "day2/day2a005000.T000.D000.P003.H001.L.vec")
    )
    assert lUnits == "pixel"
    assert vUnits == "pixel/dt"
    assert tUnits == "dt"

    # test OpenPIV vec
    lUnits, vUnits, tUnits = io.get_units(
        os.path.join(path, "openpiv/exp1_001_b.vec")
    )
    assert lUnits == "pix"


def test_load_vec():
    fname = "Run000001.T000.D000.P000.H001.L.vec"
    data = io.load_vec(os.path.join(path, "Insight", fname))
    assert data["u"].shape == (63, 63, 1)
    assert data["u"][0, 0, 0] == 0.0
    assert np.allclose(data.coords["x"][0], 0.31248)
    assert "t" in data.dims


def test_load_vc7():
    data = io.load_vc7(os.path.join(path, "VC7/2Ca.VC7"))
    assert data["u"].shape == (57, 43, 1)
    assert np.allclose(data.u.values[0, 0], -0.04354814)
    assert np.allclose(data.coords["x"][-1], 193.313795)


def test_loadopenpivtxt():
    data = io.load_vec(os.path.join(path, "openpiv", "exp1_001_b.txt"))


def test_load_directory():
    _ = pkg.resource_filename("pivpy", "data/Insight")
    data = io.load_directory(_, basename="Run*", ext=".vec")
    assert np.allclose(data["t"], [0, 1, 2, 3, 4])

    _ = pkg.resource_filename("pivpy", "data/VC7")
    data = io.load_directory(_, basename="2*", ext=".VC7")
    assert np.allclose(data["t"], [0, 1])

    data = io.load_directory(
        path=os.path.join(path, "urban_canopy"), basename="B*", ext=".vc7"
    )
    assert np.allclose(data["t"], [0, 1, 2, 3, 4])


def test_create_sample_field():
    data = io.create_sample_field(frame=3)
    assert data["t"] == 3
    data = io.create_sample_field(rows=3, cols=7)
    assert data.x.shape[0] == 7
    assert data.y.shape[0] == 3
    assert data["t"] == 0


def test_create_sample_dataset(n=3):
    data = io.create_sample_dataset(n=n)
    assert data.dims["t"] == 3
    assert np.allclose(data["t"], np.arange(3))



from pivpy import io, graphics, pivpy
import pkg_resources as pkg

# f1 = 'Run000001.T000.D000.P000.H001.L.vec'
filename = pkg.resource_filename(
    "pivpy", "data/Insight/Run000001.T000.D000.P000.H001.L.vec"
)
# load data
_d = io.load_vec(filename)


def test_showscal():
    graphics.showscal(_d, property="ke")


def test_quiver():
    graphics.quiver(_d)
    _d.piv.quiver()


def test_xarray_plot():
    _d.piv.vec2scal(property="curl")
    _d["w"].isel(t=0).plot.pcolormesh()


def test_histogram():
    graphics.histogram(_d)


def test_showf():
    graphics.showf(_d)


def test_quiver_openpiv_vec():
    filename = pkg.resource_filename("pivpy", "data/openpiv/exp1_001_b.vec")
    # load data
    _d = io.load_vec(filename)
    _d.piv.quiver()



# -*- coding: utf-8 -*-
"""
Various plots

"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation, FFMpegWriter
import xarray as xr
import os


def quiver(
    data,
    arrScale=25.0,
    threshold=None,
    nthArr=1,
    aspectratio="equal",
    colbar=False,
    colbar_orient="vertical",
    units=None,
    streamlines=False,
):
    """
    Generates a quiver plot of a 'data' xarray DataArray object (single frame
    from a dataset)
    Inputs:
        data - xarray DataArray of the type defined in pivpy, one of the
        frames in the Dataset
            selected by default using .isel(t=0)
        threshold - values above the threshold will be set equal to threshold
        arrScale - use to change arrow scales
        nthArr - use to plot only every nth arrow from the array
        contourLevels - use to specify the maximum value (abs) of contour plots
        colbar - True/False wether to generate a colorbar or not
        logscale - if true then colorbar is on log scale
        aspectratio - set auto or equal for the plot's apearence
        colbar_orient - 'horizontal' or 'vertical' orientation of the colorbar
        (if colbar is True)
    Outputs:
        none
    Usage:
        graphics.quiver(data, arrScale = 0.2, threshold = Inf, n)
    """

    data = dataset_to_array(data)

    # by construction, u and v are rows x columns so need to be rotated 90 deg
    # prior to plotting

    x = data.x.values
    y = data.y.values
    u = data.u.values
    v = data.v.values

    if u.shape[0]  == x.shape[0]:
        u = u.T
        v = v.T

    if units is not None:  # replace  units
        lUnits = units[0]  # ['m' 'm' 'mm/s' 'mm/s']
        velUnits = units[2]
        # tUnits = velUnits.split('/')[1] # make it 's' or 'dt'
    else:
        lUnits = data.attrs["units"][0]
        velUnits = data.attrs["units"][2]
        # tUnits = data.attrs['units'][2].split('/')[-1]

    # in addition, if the x,y units are pixels,
    # we should plot it in the image coordinate system
    # with 0,0 at the top left corner
    # and so v should be negative
    # and axis inverted

    # if lUnits == "pix":
    #     v = -1 * v  # only for graphics, we do not change data

    if threshold is not None:
        data["u"] = xr.where(data["u"] > threshold, threshold, data["u"])
        data["v"] = xr.where(data["v"] > threshold, threshold, data["v"])

    S = np.array(np.sqrt(u ** 2 + v ** 2))

    if len(plt.get_fignums()) == 0:  # if no figure is open
        fig, ax = plt.subplots()  # open a new figure
    else:
        fig = plt.gcf()
        ax = plt.gca()

    # quiver itself
    if colbar:
        Q = ax.quiver(
            x,
            y,
            u,
            v,
            S,
            units="width",
            scale=np.max(S * arrScale),
            headwidth=2,
        )
        cbar = fig.colorbar(Q, shrink=0.9, orientation=colbar_orient)
    else:
        ax.quiver(
            x, y, u, v, units="width", scale=np.max(S * arrScale), headwidth=2
        )

    # print(f'lUnits = {lUnits}')
    # if lUnits == "pix":
    #     ax.invert_yaxis()

    if streamlines is True:  # contours or streamlines
        speed = np.sqrt(u ** 2 + v ** 2)
        strm = ax.streamplot(
            x, y, u, v, color=speed, cmap=plt.get_cmap("hot"), linewidth=4
        )

        if colbar:
            cbar = fig.colorbar(
                strm.lines, orientation=colbar_orient, fraction=0.1
            )
            cbar.set_label(r"$ V \, (" + velUnits + r")$")

    ax.set_xlabel(f"x({lUnits})")
    ax.set_ylabel(f"y ({lUnits})")
    ax.set_aspect(aspectratio)
    # ax.invert_yaxis()

    return fig, ax


def histogram(data, normed=False):
    """
    this function will plot a normalized histogram of
    the velocity data.
    Input:
        data : xarray DataSet with ['u','v'] attrs['units']
        normed : (optional) default is False to present normalized
        histogram
    """

    u = np.asarray(data.u).flatten()
    v = np.asarray(data.v).flatten()

    units = data.attrs["units"]
    f, ax = plt.subplots(2)

    ax[0].hist(u, bins = np.int32(np.sqrt(len(u)) * 0.5), density=normed)
    ax[0].set_xlabel("u [" + units[2] + "]")

    ax[1] = plt.subplot2grid((2, 1), (1, 0))
    ax[1].hist(v, bins = np.int32(np.sqrt(len(v) * 0.5)), density=normed)
    ax[1].set_xlabel("v [" + units[2] + "]")
    plt.tight_layout()
    return f, ax


def contour_plot(
    data,
    threshold=None,
    contourLevels=None,
    colbar=None,
    logscale=False,
    aspectratio="equal",
    units=None,
):
    """ contourf ajusted for the xarray PIV dataset, creates a
        contour map for the data['w'] property.
        Input:
            data : xarray PIV DataArray, converted automatically using
            .isel(t=0)
            threshold : a threshold value, default is None (no data clipping)
            contourLevels : number of contour levels, default is None
            colbar : None (hide), 'horizontal', or 'vertical'
            logscale : boolean (True is default) create in linear/log scale
            aspectration : string, 'equal' is the default
    """

    data = dataset_to_array(data)

    if "w" not in data.var():
        data.piv.vec2scal("ke")

    if units is not None:
        lUnits = units[0]  # ['m' 'm' 'mm/s' 'mm/s']
        # velUnits = units[2]
        # tUnits = velUnits.split('/')[1] # make it 's' or 'dt'
    else:
        # lUnits, velUnits = '', ''
        lUnits = data.attrs["units"][0]
        propUnits = (
            data.attrs["variables"][-1] + data.attrs["units"][-1]
        )  # last one is from 'w'

    f, ax = plt.subplots()

    if threshold is not None:
        data["w"] = xr.where(data["w"] > threshold, threshold, data["w"])

    # m = np.amax(abs(data["w"]))
    # n = np.amin(abs(data["w"]))
    if contourLevels is None:
        levels = np.linspace(
            np.min(data["w"].values), np.max(data["w"].values), 10
        )
    else:
        levels = contourLevels  # vector of levels to set

    if logscale:
        c = ax.contourf(
            data.x,
            data.y,
            np.abs(data["w"]),
            levels=levels,
            cmap=plt.get_cmap("RdYlBu"),
            norm=plt.colors.LogNorm(),
        )
    else:
        c = ax.contourf(
            data.x,
            data.y,
            data["w"],
            levels=levels,
            cmap=plt.get_cmap("RdYlBu"),
        )

    plt.xlabel(f"x [{lUnits}]")
    plt.ylabel(f"y [{lUnits}]")
    if colbar is not None:
        cbar = plt.colorbar(c, orientation=colbar)
        cbar.set_label(propUnits)

    ax.set_aspect(aspectratio)

    return f, ax


def showf(data, property="ke", **kwargs):
    """
    showf(data, var, units)
    Arguments:
        data : xarray.DataSet that contains dimensions of t,x,y
               and variables u,v and maybe w (scalar)
    """
    data.piv.vec2scal(property=property)
    contour_plot(data)
    quiver(data, **kwargs)


def showscal(data, property="ke", **kwargs):
    """
    showf(data, var, units)
    Arguments:
        data : xarray.DataSet that contains dimensions of t,x,y
               and a variable w (scalar)
    """
    data.piv.vec2scal(property=property)
    fig, ax = contour_plot(data, **kwargs)
    return fig, ax


def animate(data, arrowscale=1, savepath=None):
    """ animates the quiver plot for the dataset (multiple frames)
    Input:
        data : xarray PIV type of DataSet
        arrowscale : [optional] integer, default is 1
        savepath : [optional] path to save the MP4 animation, default is None

    Output:
        if savepath is None, then only an image display of the animation
        if savepath is an existing path, a file named im.mp4 is saved

    """
    X, Y = np.meshgrid(data.x, data.y)
    X = X.T
    Y = Y.T
    U, V = data.u[:, :, 0], data.v[:, :, 0]  # first frame
    fig, ax = plt.subplots(1, 1)
    M = np.sqrt(U ** 2 + V ** 2)

    Q = ax.quiver(
        X[::3, ::3],
        Y[::3, ::3],
        U[::3, ::3],
        V[::3, ::3],
        M[::3, ::3],
        units="inches",
        scale=arrowscale,
    )

    cb = plt.colorbar(Q)

    units = data.attrs["units"]

    cb.ax.set_ylabel("velocity (" + units[2] + ")")

    text = ax.text(
        0.2,
        1.05,
        "1/" + str(len(data.t)),
        ha="center",
        va="center",
        transform=ax.transAxes,
    )

    def update_quiver(num, Q, data, text):
        U, V = data.u[:, :, num], data.v[:, :, num]

        M = np.sqrt(U[::3, ::3] ** 2 + V[::3, ::3] ** 2)
        Q.set_UVC(U[::3, ::3], V[::3, ::3], M[::3, ::3])
        text.set_text(str(num + 1) + "/" + str(len(data.t)))
        return Q

    anim = FuncAnimation(
        fig,
        update_quiver,
        fargs=(Q, data, text),
        frames=len(data.t),
        blit=False,
    )
    mywriter = FFMpegWriter()
    if savepath:
        p = os.getcwd()
        os.chdir(savepath)
        anim.save("im.mp4", writer=mywriter)
        os.chdir(p)
    else:
        anim.save("im.mp4", writer=mywriter)


def dataset_to_array(data, t=0):
    """ converts xarray Dataset to array """
    if "t" in data.dims:
        print("Warning: function for a single frame, using the first \
               frame, supply data.isel(t=N)")
        data = data.isel(t=t)

    if "z" in data.dims:
        print("Warning: using first z cordinate, use data.isel(z=0)")
        data = data.isel(z=0)
    return data



import os
from setuptools import setup, find_packages

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="pivpy",
    version="0.0.10",
    author="Turbulence Structure Laboratory, Tel Aviv University",
    author_email="alex.liberzon@gmail.com",
    description=(
        "Post processing of PIV vector files "
        " eventually going into OpenPIV "
    ),
    license="BSD",
    keywords="example documentation tutorial",
    url="http://packages.python.org/pivpy",
    packages=find_packages(),
    include_package_data=True,
    test_suite="tests",
    install_requires=[
        "numpy",
        "scipy",
        "xarray",
        "matplotlib",
        "pytest",
    ],
    long_description=read("README.md"),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: BSD License",
    ],
)



# -*- coding: utf-8 -*-

"""
Contains functions for reading flow fields in various formats
"""

import numpy as np
import xarray as xr
from glob import glob
import os
import re
import ReadIM

default_units = ["pix", "pix", "pix/dt", "pix/dt"]
default_variables = ["x", "y", "u", "v", "s2n"]


def create_sample_field(rows=5, cols=8, frame=0, noise_sigma=1.0):
    """ creates a sample dataset for the tests """

    x = np.arange(32.0, (cols + 1) * 32.0, 32.0)
    y = np.arange(16.0, (rows + 1) * 16.0, 16.0)

    xm, ym = np.meshgrid(x, y)
    u = np.ones_like(xm) + np.linspace(0.0, 10.0, cols)
    v = (
        np.zeros_like(ym)
        + np.linspace(0.0, 1.0, rows).reshape(rows, 1)
        + noise_sigma * np.random.randn(rows, 1)
    )

    u = u[:, :, np.newaxis]
    v = v[:, :, np.newaxis]
    chc = np.ones_like(u)

    u = xr.DataArray(
        u, dims=("y", "x", "t"), coords={"x": x, "y": y, "t": [frame]}
    )
    v = xr.DataArray(
        v, dims=("y", "x", "t"), coords={"x": x, "y": y, "t": [frame]}
    )
    chc = xr.DataArray(
        chc, dims=("y", "x", "t"), coords={"x": x, "y": y, "t": [frame]}
    )

    data = xr.Dataset({"u": u, "v": v, "chc": chc})

    data.attrs["variables"] = ["x", "y", "u", "v"]
    data.attrs["units"] = ["pix", "pix", "pix/dt", "pix/dt"]
    data.attrs["dt"] = 1.0
    data.attrs["files"] = ""

    return data


def create_sample_dataset(n=5):
    """ using create_sample_field that has random part in it, create
    a sample dataset of length 'n' """

    data = []
    for i in range(n):
        data.append(create_sample_field(frame=i))

    combined = xr.concat(data, dim="t")
    combined.attrs["variables"] = ["x", "y", "u", "v"]
    combined.attrs["units"] = ["pix", "pix", "pix/dt", "pix/dt"]
    combined.attrs["dt"] = 1.0
    combined.attrs["files"] = ""

    return combined

def create_uniform_strain():
    return create_sample_field(noise_sigma=0.0)

    
def from_arrays(x, y, u, v, mask):
    """
        from_arrays(x,y,u,v,mask,frame=0)
        creates an xArray Dataset from 5 two-dimensional Numpy arrays
        of x,y,u,v and mask

        Input:
            x,y,u,v,mask = Numpy floating arrays, all the same size
        Output:
            data is a xAarray Dataset, see xarray for help
    """
    # create data structure of appropriate size
    data = create_sample_field(rows=x.shape[0], cols=x.shape[1])
    # assign arrays
    data["x"] = x[0, :]
    data["y"] = y[:, 0]
    data["u"] = xr.DataArray(u.T[:, :, np.newaxis], dims=("x", "y", "t"))
    data["v"] = xr.DataArray(v.T[:, :, np.newaxis], dims=("x", "y", "t"))
    data["chc"] = xr.DataArray(mask.T[:, :, np.newaxis], dims=("x", "y", "t"))

    return data

def from_df(df, frame=0, dt=1.0, filename=''):
    """
        from_df(x,y,u,v,mask,frame=0)
        creates an xArray Dataset from pandas dataframe with 5 columns

        Read the .txt files faster with pandas read_csv()

        %%time
        df = pd.read_csv(files_list[-1],delimiter='\t', 
                        names = ['x','y','u','v','mask'],header=0)
        from_df(df,filename=files_list[-1])

        is 3 times faster than the load_txt


        Input:
            x,y,u,v,mask = Numpy floating arrays, all the same size
        Output:
            data is a xAarray Dataset, see xarray for help
    """
    d = df.to_numpy()

    x = np.unique(d[:, 0])
    y = np.unique(d[:, 1])
    d = d.reshape(len(y), len(x), 5)  # .transpose(1, 0, 2)

    u = d[:, :, 2]
    v = d[:, :, 3]
    chc = d[:, :, 4]

    # extend dimensions
    u = u[:, :, np.newaxis]
    v = v[:, :, np.newaxis]
    chc = chc[:, :, np.newaxis]


    u = xr.DataArray(
        u, dims=("y", "x", "t"), coords={"x": x, "y": y, "t": [frame]}
    )
    v = xr.DataArray(
        v, dims=("y", "x", "t"), coords={"x": x, "y": y, "t": [frame]}
    )
    chc = xr.DataArray(
        chc, dims=("y", "x", "t"), coords={"x": x, "y": y, "t": [frame]}
    )

    data = xr.Dataset({"u": u, "v": v, "chc": chc})

    data.attrs["variables"] = df.columns.to_list()
    data.attrs["units"] = ['pix','pix','pix/dt','pix/dt']
    data.attrs["dt"] = dt
    data.attrs["files"] = filename

    return data

def load_vec(
    filename,
    rows=None,
    cols=None,
    variables=default_variables,
    units=default_units,
    dt=1.0,
    frame=0,
):
    """
        load_vec(filename,rows=rows,cols=cols)
        Loads the VEC file (TECPLOT format by TSI Inc.),
        OpenPIV VEC or TXT formats
        Arguments:
            filename : file name, expected to have a header and 5 columns
            rows, cols : number of rows and columns of a vector field,
            if None, None, then parse_header is called to infer the number
            written in the header
            dt : time interval (default is None)
            frame : frame or time marker (default is None)
        Output:
            data is a xAarray Dataset, see xarray for help
    """
    if rows is None or cols is None:
        variables, units, rows, cols, dt, frame = parse_header(filename)

    if rows is None:  # means no headers
        d = np.genfromtxt(filename, usecols=(0, 1, 2, 3, 4))
        x = unique(d[:, 0])
        y = unique(d[:, 1])
        d = d.reshape(len(y), len(x), 5)  # .transpose(1, 0, 2)
    else:
        # d = np.genfromtxt(
        #     filename, skiprows=1, delimiter=",", usecols=(0, 1, 2, 3, 4)
        # ).reshape(rows, cols, 5)
        d = np.genfromtxt(
            filename, skip_header=1, delimiter=",", usecols=(0, 1, 2, 3, 4)
        ).reshape(cols, rows, 5)
        x = d[:, :, 0][0, :]
        y = d[:, :, 1][:, 0]

    u = d[:, :, 2]
    v = d[:, :, 3]
    chc = d[:, :, 4]

    # extend dimensions
    u = u[:, :, np.newaxis]
    v = v[:, :, np.newaxis]
    chc = chc[:, :, np.newaxis]

    u = xr.DataArray(
        u, dims=("y", "x", "t"), coords={"x": x, "y": y, "t": [frame]}
    )
    v = xr.DataArray(
        v, dims=("y", "x", "t"), coords={"x": x, "y": y, "t": [frame]}
    )
    chc = xr.DataArray(
        chc, dims=("y", "x", "t"), coords={"x": x, "y": y, "t": [frame]}
    )

    data = xr.Dataset({"u": u, "v": v, "chc": chc})

    data.attrs["variables"] = variables
    data.attrs["units"] = units
    data.attrs["dt"] = dt
    data.attrs["files"] = filename

    return data


def load_directory(path, basename="*", ext=".vec"):
    """
    load_directory (path,basename='*', ext='*.vec')

    Loads all the files with the chosen sextension in the directory into a
    single xarray Dataset with variables and units added as attributes

    Input:
        directory : path to the directory with .vec, .txt or .VC7 files,
        period . can be dropped

    Output:
        data : xarray DataSet with dimensions: x,y,t and
               data arrays of u,v,
               attributes of variables and units


    See more: load_vec
    """
    files = sorted(glob(os.path.join(path, basename + ext)))
    if len(files) == 0:
        raise IOError(f"No files {basename+ext} in the directory {path} ")
    else:
        print(f"found {len(files)} files")

    data = []
    combined = []

    if ext.lower().endswith("vec"):
        variables, units, rows, cols, dt, frame = parse_header(files[0])

        for i, f in enumerate(files):
            data.append(
                load_vec(f, rows, cols, variables, units, dt, frame + i - 1)
            )

        if len(data) > 0:
            combined = xr.concat(data, dim="t")
            combined.attrs["variables"] = data[0].attrs["variables"]
            combined.attrs["units"] = data[0].attrs["units"]
            combined.attrs["dt"] = data[0].attrs["dt"]
            combined.attrs["files"] = files
    elif ext.lower().endswith("vc7"):
        frame = 1
        for i, f in enumerate(files):
            if basename == "B*":  # quite strange to have a specific name?
                time = int(f[-9:-4]) - 1
            else:
                time = i
            data.append(load_vc7(f, time))
        if len(data) > 0:
            combined = xr.concat(data, dim="t")
            combined.attrs = data[-1].attrs
    elif ext.lower().endswith("txt"):
        variables, units, rows, cols, dt, frame = parse_header(files[0])

        for i, f in enumerate(files):
            data.append(
                load_txt(f, rows, cols, variables, units, dt, frame + i - 1)
            )
        if len(data) > 0:
            combined = xr.concat(data, dim="t")
            combined.attrs["variables"] = data[0].attrs["variables"]
            combined.attrs["units"] = data[0].attrs["units"]
            combined.attrs["dt"] = data[0].attrs["dt"]
            combined.attrs["files"] = files
        else:
            raise IOError("Could not read the files")

    return combined


def parse_header(filename):
    """
    parse_header ( filename)
    Parses header of the file (.vec) to get the variables (typically X,Y,U,V)
    and units (can be m,mm, pix/dt or mm/sec, etc.), and the size of the
    dataset by the number of rows and columns.
    Input:
        filename : complete path of the file to read
    Returns:
        variables : list of strings
        units : list of strings
        rows : number of rows of the dataset
        cols : number of columns of the dataset
        dt   : time interval between the two PIV frames in microseconds
    """

    # defaults
    frame = 0

    # split path from the filename
    fname = os.path.basename(filename)
    # get the number in a filename if it's a .vec file from Insight
    if "." in fname[:-4]:  # day2a005003.T000.D000.P003.H001.L.vec
        frame = int(re.findall(r"\d+", fname.split(".")[0])[-1])
    elif "_" in filename[:-4]:
        frame = int(
            re.findall(r"\d+", fname.split("_")[1])[-1]
        )  # exp1_001_b.vec, .txt

    with open(filename) as fid:
        header = fid.readline()

    # if the file does not have a header, can be from OpenPIV or elsewhere
    # return None
    if header[:5] != "TITLE":
        return (
            ["x", "y", "u", "v"],
            ["pix", "pix", "pix/dt", "pix/dt"],
            None,
            None,
            None,
            frame,
        )

    header_list = (
        header.replace(",", " ").replace("=", " ").replace('"', " ").split()
    )

    # get variable names, typically X,Y,U,V
    variables = header_list[3:12][::2]

    # get units - this is important if it's mm or m/s
    units = header_list[4:12][::2]

    # get the size of the PIV grid in rows x cols
    rows = int(header_list[-5])
    cols = int(header_list[-3])

    # this is also important to know the time interval, dt
    ind1 = header.find("MicrosecondsPerDeltaT")
    dt = float(header[ind1:].split('"')[1])

    return (variables, units, rows, cols, dt, frame)


def get_units(filename):
    """
    get_units(filename)

    given a full path name to the .vec file will return the names
    of length and velocity units fallback option is all None. Uses
    parse_header function, see below.

    """

    # lUnits, velUnits, tUnits = 'pixel', 'pixel', 'dt'

    _, units, _, _, _, _ = parse_header(filename)

    if units == "":
        return "pix", "pix", "dt"

    lUnits = units[0]
    velUnits = units[2]

    if velUnits == "pixel":
        velUnits = velUnits + "/dt"  # make it similar to m/s

    tUnits = velUnits.split("/")[1]  # make it 's' or 'dt'

    return lUnits, velUnits, tUnits


def load_vc7(path, time=0):
    """
    input path for files format from davis tested for im7&vc7
    out put [X Y U V mask]
    valid only for 2d piv cases
    RETURN:
     in case of images (image type=0):
              X = scaled x-coordinates
              Y = scaled y-coordinates
              U = scaled image intensities
              v=0
              MASK=0
            in case of 2D vector fields (A.IType = 1,2 or 3):
              X = scaled x-coordinates
               Y = scaled y-coordinates
               U = scaled vx-components of vectors
               V = scaled vy-components of vectors

    """
    # you need to add clear to prevent data leaks
    buff, vatts = ReadIM.extra.get_Buffer_andAttributeList(path)
    v_array, buff1 = ReadIM.extra.buffer_as_array(buff)
    nx = buff.nx
    # nz = buff.nz # flake8 claims it's not used
    ny = buff.ny
    # set data range:
    baseRangeX = np.arange(nx)
    baseRangeY = np.arange(ny)
    # baseRangeZ = np.arange(nz)
    lhs1 = (
        baseRangeX + 0.5
    ) * buff.vectorGrid * buff.scaleX.factor + buff.scaleX.offset  # x-range
    lhs2 = (
        baseRangeY + 0.5
    ) * buff.vectorGrid * buff.scaleY.factor + buff.scaleY.offset  # y-range
    lhs3 = 0
    lhs4 = 0
    mask = 0
    if buff.image_sub_type <= 0:  # grayvalue image format
        [lhs1, lhs2] = np.meshgrid(lhs1, lhs2)
        lhs3 = v_array[0, :, :]
        lhs4 = v_array[1, :, :]
        Im = xr.DataArray(
            v_array,
            dims=("frame", "z", "x"),
            coords={"x": lhs1[0, :], "z": lhs2[:, 0], "frame": [0, 1]},
        )
        data = xr.Dataset({"Im": Im})

    elif buff.image_sub_type == 2:  # simple 2D vector format: (vx,vy)
        # Calculate vector position and components
        [lhs1, lhs2] = np.meshgrid(lhs1, lhs2)
        #    lhs1=np.transpose(lhs1)
        #    lhs2=np.transpose(lhs2)
        lhs3 = v_array[0, :, :] * buff.scaleI.factor + buff.scaleI.offset
        lhs4 = v_array[1, :, :] * buff.scaleI.factor + buff.scaleI.offset
        if buff.scaleY.factor < 0.0:
            lhs4 = -lhs4
        lhs3 = lhs3[:, :, np.newaxis]
        lhs4 = lhs4[:, :, np.newaxis]
        u = xr.DataArray(
            lhs3,
            dims=("z", "x", "t"),
            coords={"x": lhs1[0, :], "z": lhs2[:, 0], "t": [time]},
        )
        v = xr.DataArray(
            lhs4,
            dims=("z", "x", "t"),
            coords={"x": lhs1[0, :], "z": lhs2[:, 0], "t": [time]},
        )
        data = xr.Dataset({"u": u, "v": v})
    #    	plt.quiver(lhs1,lhs2,lhs3,lhs4);
    elif buff.image_sub_type == 3 or buff.image_sub_type == 1:
        # normal 2D vector format + peak: sel+4*(vx,vy) (+peak)
        # Calculate vector position and components
        [lhs1, lhs2] = np.meshgrid(lhs1, lhs2)
        #    lhs1=np.transpose(lhs1)
        #    lhs2=np.transpose(lhs2)
        lhs3 = lhs1 * 0
        lhs4 = lhs2 * 0
        # Get choice
        maskData = v_array[0, :, :].astype(np.int8)
        # Build best vectors from choice field
        for i in range(5):
            mask = maskData == (i + 1)
            if i < 4:  # get best vectors
                dat = v_array[2 * i + 1, :, :]
                lhs3[mask] = dat[mask]
                dat = v_array[2 * i + 2, :, :]
                lhs4[mask] = dat[mask]
            else:  # get interpolated vectors
                dat = v_array[7, :, :]
                lhs3[mask] = dat[mask]
                dat = v_array[8, :, :]
                lhs4[mask] = dat[mask]
        lhs3 = lhs3 * buff.scaleI.factor + buff.scaleI.offset
        lhs4 = lhs4 * buff.scaleI.factor + buff.scaleI.offset
        # Display vector field
        if buff.scaleY.factor < 0.0:
            lhs4 = -1 * lhs4
        lhs3 = lhs3.T[:, :, np.newaxis]
        lhs4 = lhs4.T[:, :, np.newaxis]
        chc = maskData.T[:, :, np.newaxis]
        u = xr.DataArray(
            lhs3,
            dims=("x", "y", "t"),
            coords={"x": lhs1[0, :], "y": lhs2[:, 0], "t": [time]},
        )
        v = xr.DataArray(
            lhs4,
            dims=("x", "y", "t"),
            coords={"x": lhs1[0, :], "y": lhs2[:, 0], "t": [time]},
        )
        chc = xr.DataArray(
            chc,
            dims=("x", "y", "t"),
            coords={"x": lhs1[0, :], "y": lhs2[:, 0], "t": [time]},
        )
        data = xr.Dataset({"u": u, "v": v, "chc": chc})
    if buff.image_sub_type > 0:
        data.attrs = ReadIM.extra.att2dict(vatts)
        data.attrs["variables"] = ["x", "y", "u", "v"]
        data.attrs["units"] = ["mm", "mm", "m/s", "m/s"]
        data.attrs["dt"] = int(data.attrs["FrameDt0"][:-3])
        data.attrs["files"] = path
    # clean memory
    ReadIM.DestroyBuffer(buff1)
    del buff1
    ReadIM.DestroyBuffer(buff)
    del buff
    ReadIM.DestroyAttributeListSafe(vatts)
    del vatts
    return data


def load_txt(
    filename,
    rows=None,
    cols=None,
    variables=default_variables,
    units=default_units,
    dt=1.0,
    frame=0,
):
    """
        load_vec(filename,rows=rows,cols=cols)
        Loads the VEC file (TECPLOT format by TSI Inc.), OpenPIV VEC or TXT
        formats
        Arguments:
            filename : file name, expected to have a header and 5 columns
            rows, cols : number of rows and columns of a vector field,
            if None, None, then parse_header is called to infer the number
            written in the header
            dt : time interval (default is None)
            frame : frame or time marker (default is None)
        Output:
            data is a xAarray Dataset, see xarray for help
    """
    if rows is None:  # means no headers
        d = np.genfromtxt(filename, usecols=(0, 1, 2, 3, 4))
        x = unique(d[:, 0])
        y = unique(d[:, 1])
        d = d.reshape(len(y), len(x), 5).transpose(1, 0, 2)
    else:
        d = np.genfromtxt(
            filename, skip_header=1, delimiter=",", usecols=(0, 1, 2, 3, 4)
        ).reshape(rows, cols, 5)
        x = d[:, :, 0][0, :]
        y = d[:, :, 1][:, 0]

    u = d[:, :, 2]
    v = d[:, :, 3]
    chc = d[:, :, 4]

    # extend dimensions
    u = u[:, :, np.newaxis]
    v = v[:, :, np.newaxis]
    chc = chc[:, :, np.newaxis]

    u = xr.DataArray(
        u, dims=("x", "y", "t"), coords={"x": x, "y": y, "t": [frame]}
    )
    v = xr.DataArray(
        v, dims=("x", "y", "t"), coords={"x": x, "y": y, "t": [frame]}
    )
    chc = xr.DataArray(
        chc, dims=("x", "y", "t"), coords={"x": x, "y": y, "t": [frame]}
    )

    data = xr.Dataset({"u": u, "v": v, "chc": chc})

    data.attrs["variables"] = variables
    data.attrs["units"] = units
    data.attrs["dt"] = dt
    data.attrs["files"] = filename

    return data


def unique(array):
    """ Returns not sorted unique """
    uniq, index = np.unique(array, return_index=True)
    return uniq[index.argsort()]
